/* 
Copyright 2021 Magnus Wirsig

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
SimpleFEC: simple forward error correction with small footprint
This file contains all functions so it can be used as a library.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <inttypes.h>
#include <string.h>
#include "SimpleFec.h"


#ifdef AVR
#include <avr/pgmspace.h>
const __flash
#endif
uint32_t syndrome[] = {
    0x02AAAD5B,
    0x00CDB66D,
    0x03C3C78E,
    0x03FC07F0,
    0x03FFF800
};


uint8_t bitCount(uint32_t data)
{
    uint8_t count = 0;
    for(int i=0; i<32; i++)
	if(data & ((uint32_t)1<<i))
	    count++;
    return count;
}

uint8_t calcParity(uint32_t data)
{

#ifdef AVR
    uint32_t syndrome1[5];
    memcpy_P(syndrome1, syndrome, sizeof(syndrome));
#else
	uint32_t *syndrome1 = syndrome;
#endif
    uint8_t par;
    par = bitCount(data & syndrome1[0]) % 2;
    par |= (bitCount(data & syndrome1[1]) % 2) << 1;
    par |= (bitCount(data & syndrome1[2]) % 2) << 2;
    par |= (bitCount(data & syndrome1[3]) % 2) << 3;
    par |= (bitCount(data & syndrome1[4]) % 2) << 4;
    return(par);
}

int repairData(uint32_t *data, uint8_t par)
{
    uint8_t parDiff = bitCount(calcParity(*data) ^ par);

    // No difference or only parity bit corrupt
    if(parDiff<2)
	return 0;

    for(int i = 0; i<26; i++)
    {
	uint32_t testData = (*data) ^ ((uint32_t)1<<i);
	if( calcParity(testData) ==  par)
	{
	    (*data) = testData;

	    return 1;
	}
    }
    return (-1);
}


uint32_t Get26bitBlock(void *baseAddr, uint16_t i)
{
    uint64_t tmp;
    tmp = *((uint64_t*)(baseAddr + (i*26)/8));
    tmp = tmp >> ((i*26)%8);
    tmp &= 0x03FFFFFFUL;
    return tmp;
}

void Set26bitBlock(void *baseAddr, uint16_t i, uint32_t cont)
{
    uint64_t tmp;
    cont &= (uint64_t)0x03FFFFFF;
    tmp = *((uint64_t*)(baseAddr + (i*26)/8));
    tmp &= ~( (uint64_t)0x03FFFFFFULL << ((i*26)%8));
    tmp |= cont << ((i*26)%8);
    *((uint64_t*)(baseAddr + (i*26)/8)) = tmp;
}

uint8_t Get5bitBlock(void *baseAddr, uint16_t i)
{
    uint16_t tmp;
    tmp = *((uint16_t*)(baseAddr + (i*5)/8));
    tmp = tmp >> ((i*5)%8);
    tmp &= 0x1F;
    return tmp;
}

void Set5bitBlock(void *baseAddr, uint16_t count, uint8_t content)
{
    for(int i = 0; i<5; i++)
    {
	uint16_t bitAddr = count*5 + i;

	if(content&(1<<i))
	    *((uint8_t*)(baseAddr+bitAddr/8)) |= 1<<bitAddr%8;
	else
	    *((uint8_t*)(baseAddr+bitAddr/8)) &= ~(1<<bitAddr%8);
    }
}


void SimpleFecPutFec(void *data, int len, void *fecData)
{
    uint32_t tmpData;
    uint8_t tmpFec;
    uint16_t fecLen = ((len*8)/26) + (((len*8)%26)>0);

    for(int i=0; i < fecLen; i++)
    {
	tmpData = Get26bitBlock(data, i);
	tmpFec = calcParity(tmpData);
	Set5bitBlock(fecData, i, tmpFec);
    }
}

uint16_t SimpleFecGetFecDataLength(int len)
{
    uint16_t blockCount26 = ((len*8)/26);
    if((len*8)%26)
	blockCount26++;

    uint16_t blockCount5 = blockCount26;
    uint16_t blockCount8 = (blockCount5*5)/8;
	if( (blockCount5*5)%8)
             blockCount8++;

    return blockCount8;
}


void SimpleFecRepairSingleBlock(void *data, uint16_t pos, void *fecData)
{
    uint32_t tmpData;
    uint8_t tmpFec;
    tmpData = Get26bitBlock(data, pos);
    tmpFec = Get5bitBlock(fecData, pos);
    if(repairData(&tmpData, tmpFec) == 1)
	Set26bitBlock(data, pos, tmpData); 
}


void SimpleFecRepairData(void *data, int len, void *fecData)
{
    uint32_t tmpData;
    uint8_t tmpFec;
    int repaired = 0;
    int failed = 0;
    uint16_t fecLen = ((len*8)/26) + (((len*8)%26)>0);

    for(int i=0; i<fecLen; i++)
    {
	tmpData = Get26bitBlock(data, i);
	tmpFec = Get5bitBlock(fecData, i);
	switch (repairData(&tmpData, tmpFec))
	{
	case -1: failed++; break;
	case 1: repaired++; Set26bitBlock(data, i, tmpData); break;
	}
    }
}

uint8_t CalcXorChecksum(void *data, uint16_t length)
{
    uint8_t checksum = 0;
    for(uint16_t i=0; i<length; i++)
	checksum ^= *((uint8_t*)data+i);
    return(checksum);
}
